from __future__ import annotations

from dataclasses import (
    fields,
    is_dataclass,
    replace,
)
from typing import (
    Protocol,
    TypeVar,
    TYPE_CHECKING,
)

if TYPE_CHECKING:
    from _typeshed import DataclassInstance as _DataclassInstance

    class _HasName(Protocol, _DataclassInstance):
        name: str

    _DataclassT = TypeVar('_DataclassT', bound=_DataclassInstance)
    _HasNameT = TypeVar('_HasNameT', bound=_HasName)


def default(
    cls: type[_DataclassT],
    /,
) -> _DataclassT:
    f = {}

    for field in fields(cls):
        if 'default' in field.metadata:
            f[field.name] = field.metadata['default']

    return cls(**f)


def merge(
    self: _DataclassT,
    other: _DataclassT | None, /,
) -> _DataclassT:
    if other is None:
        return self

    f = {}

    for field in fields(self):
        if not field.init:
            continue

        field_default_type = object
        if isinstance(field.default_factory, type):
            field_default_type = field.default_factory

        self_field = getattr(self, field.name)
        other_field = getattr(other, field.name)

        if field.name == 'name':
            assert self_field == other_field
        elif field.type == 'bool':
            f[field.name] = other_field
        elif field.metadata.get('merge') == 'assoclist':
            f[field.name] = _merge_assoclist(self_field, other_field)
        elif is_dataclass(field_default_type):
            f[field.name] = merge(self_field, other_field)
        elif issubclass(field_default_type, list):
            f[field.name] = self_field + other_field
        elif issubclass(field_default_type, dict):
            f[field.name] = self_field | other_field
        elif field.default is None:
            if other_field is not None:
                f[field.name] = other_field
        else:
            raise RuntimeError(f'Unable to merge for type {field.type}')

    return replace(self, **f)


def merge_default(
    cls: type[_DataclassT],
    /,
    *others: _DataclassT,
) -> _DataclassT:
    ret: _DataclassT = default(cls)
    for o in others:
        ret = merge(ret, o)
    return ret


def _merge_assoclist(
    self_list: list[_HasNameT],
    other_list: list[_HasNameT],
    /,
) -> list[_HasNameT]:
    '''
    Merge lists where each item got a "name" attribute
    '''
    if not self_list:
        return other_list
    if not other_list:
        return self_list

    ret: list[_HasNameT] = []
    other_dict = {
        i.name: i
        for i in other_list
    }
    for i in self_list:
        if i.name in other_dict:
            ret.append(merge(i, other_dict.pop(i.name)))
        else:
            ret.append(i)
    ret.extend(other_dict.values())
    return ret



import dataclasses
import pathlib
import re
import tomllib
from typing import Optional, Self

import dacite

from debian_linux.debian import PackageRelation


_dacite_config = dacite.Config(
    cast=[PackageRelation],
    strict=True,
)


@dataclasses.dataclass
class ConfigBase:
    uri: str
    links_excluded: list[str]
    files_unpackaged: list[str]


@dataclasses.dataclass
class ConfigEula:
    title: str
    text: str


@dataclasses.dataclass
class ConfigPackage:
    name: str
    desc: str
    longdesc: str
    support: list[str] = dataclasses.field(default_factory=list)
    suggests: PackageRelation = dataclasses.field(default_factory=PackageRelation)
    recommends: PackageRelation = dataclasses.field(default_factory=PackageRelation)
    depends: PackageRelation = dataclasses.field(default_factory=PackageRelation)
    conflicts: PackageRelation = dataclasses.field(default_factory=PackageRelation)
    breaks: PackageRelation = dataclasses.field(default_factory=PackageRelation)
    replaces: PackageRelation = dataclasses.field(default_factory=PackageRelation)
    provides: PackageRelation = dataclasses.field(default_factory=PackageRelation)
    eula: Optional[ConfigEula] = None
    uri: Optional[str] = None
    files: list[str] = dataclasses.field(default_factory=list)
    files_excluded: list[str] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class Config:
    # Disable basic fields
    name: str = dataclasses.field(init=False, repr=False, default='')
    enable: bool = dataclasses.field(init=False, repr=False, default=True)

    base: ConfigBase
    package: list[ConfigPackage]

    @classmethod
    def read(cls) -> Self:
        with open('debian/config/defines.toml', 'rb') as f:
            data = tomllib.load(f)

        return dacite.from_dict(
            data_class=cls,
            data=data,
            config=_dacite_config,
        )


_wildcard_re = re.compile(r'\*\*/?|[*?.^$+{}\\\[\]|()]')
_wildcard_map = {
    '**/': r'(?:.+/)?',
    '**':  r'.*',
    '*':   r'[^/]*',
    '?':   r'[^/]',
}

# pathlib.Path.match() does *not* support '**', so do our own
# pattern-to-regexp conversion
def pattern_to_re(pattern):
    return re.compile(
        _wildcard_re.sub(
            lambda match: _wildcard_map.get(match.group(0),
                                            '\\' + match.group(0)),
            pattern))



#!/usr/bin/env python3

import dataclasses
import io
import locale
import os
import pathlib
import re
import sys
from typing import Iterable, Optional

locale.setlocale(locale.LC_CTYPE, "C.UTF-8")

from debian_firmware.config import Config, pattern_to_re
from debian_linux.dataclasses_deb822 import field_deb822, read_deb822, write_deb822
from debian_linux.debian import BinaryPackage as BinaryPackageBase, PackageDescription, PackageRelation
import debian_linux.gencontrol
from debian_linux.gencontrol import MakeFlags
from debian_linux.utils import Templates as TemplatesBase


# XXX Delete after this field is added in linux-support
@dataclasses.dataclass
class BinaryPackage(BinaryPackageBase):
    homepage: 'Optional[str]' = field_deb822(
        'Homepage',
        default=None,
    )


@dataclasses.dataclass
class Template:
    template: 'str' = field_deb822('Template')
    type: 'str' = field_deb822('Type')
    default: 'Optional[str]' = field_deb822(
        'Default',
        default=None,
    )
    description: PackageDescription = field_deb822(
        'Description',
        default_factory=PackageDescription,
    )


class Templates(TemplatesBase):
    def get_control(
        self, key: str, context: dict[str, str] = {},
    ) -> Iterable[BinaryPackage]:
        return read_deb822(BinaryPackage, io.StringIO(self.get(key, context)))

    def get_templates_control(
        self, key: str, context: dict[str, str] = {}
    ) -> Iterable[Template]:
        return read_deb822(Template, io.StringIO(self.get(key, context)))


class GenControl(debian_linux.gencontrol.Gencontrol):
    def __init__(self):
        super().__init__(Config.read(), Templates())

    def do_main(self):
        for config_entry in self.config.package:
            self.do_package(config_entry)

    def do_package(self, config_entry):
        package = config_entry.name
        vars = {}
        for field_name in ['desc', 'longdesc',
                           'suggests','recommends', 'depends',
                           'conflicts', 'breaks', 'replaces', 'provides']:
            field_value = getattr(config_entry, field_name)
            vars[field_name] = str(field_value) if field_value else ''
        vars['uri'] = (config_entry.uri
                       if config_entry.uri is not None
                       else self.config.base.uri)
        vars['package'] = package
        vars['package_env_prefix'] = 'FIRMWARE_' + package.upper().replace('-', '_')

        try:
            os.unlink('debian/firmware-%s.bug-presubj' % package)
        except OSError:
            pass
        os.symlink('bug-presubj', 'debian/firmware-%s.bug-presubj' % package)

        packages_binary = list(self.templates.get_control("binary.control", vars))

        scripts = {}

        if 'update-initramfs' in config_entry.support:
            scripts.setdefault("postinst", []).append(
                self.templates.get('postinst.update-initramfs', vars))

        if config_entry.eula:
            vars['license_title'] = config_entry.eula.title

            scripts.setdefault("preinst", []).append(
                self.templates.get('preinst.license', vars))

            templates = list(self.templates.get_templates_control('templates.license', vars))
            templates[0].description.append(re.sub('\n\n', '\n.\n', config_entry.eula.text))
            templates_filename = "debian/firmware-%s.templates" % package
            with open(templates_filename, 'w') as templates_fh:
                write_deb822(templates, templates_fh)

            desc = packages_binary[0].description
            desc.append(
"""This firmware is covered by the %s.
You must agree to the terms of this license before it is installed."""
% config_entry.eula.title)
            packages_binary[0].pre_depends = PackageRelation('debconf | debconf-2.0')

        for script, script_contents in scripts.items():
            script_contents.insert(0, "#!/bin/sh\n\nset -e\n")
            script_contents.append("#DEBHELPER#\n\nexit 0\n")
            with open("debian/firmware-%s.%s" % (package, script), "w") as script_fh:
                script_fh.write("\n".join(script_contents))

        self.bundle.add_packages(packages_binary, (package,), MakeFlags())

    def write(self):
        # We don't want to generate a makefile
        self.bundle.write_control()


if __name__ == '__main__':
    GenControl()()



#!/usr/bin/env python3

import errno, filecmp, fnmatch, glob, pathlib, re, sys
from debian import deb822
from enum import Enum

from debian_firmware.firmware import FirmwareWhence
from debian_firmware.config import Config, pattern_to_re

class DistState(Enum):
    undistributable = 1
    non_free = 2
    free = 3

def is_source_available(section):
    for file_info in section.files.values():
        if not (file_info.source
                or file_info.binary.endswith('.txt')
                or file_info.binary.endswith('.cis')):
            return False
    return True

def check_section(section):
    if section.licence is None:
        # Maybe undistributable
        return DistState.undistributable
    elif re.search(r'^BSD\b'
                   r'|^GPLv2 or OpenIB\.org BSD\b'
                   r'|^Apache-2\.0\b'
                   r'|^MIT\b'
                   r'|\bPermission\s+is\s+hereby\s+granted\s+for\s+the\s+'
                   r'distribution\s+of\s+this\s+firmware\s+(?:data|image)\b'
                   r'(?!\s+as\s+part\s+of)'
                   r'|\bRedistribution\s+and\s+use\s+in(?:\s+source\s+and)?'
                   r'\s+binary\s+forms\b'
                   r'|\bPermission\s+is\s+hereby\s+granted\b[^.]+\sto'
                   r'\s+deal\s+in\s+the\s+Software\s+without'
                   r'\s+restriction\b'
                   r'|\bredistributable\s+in\s+binary\s+form\b'
                   r'|\bgrants\s+permission\s+to\s+use\s+and\s+redistribute'
                   r'\s+these\s+firmware\s+files\b',
                   section.licence):
        return (DistState.free if is_source_available(section)
                else DistState.non_free)
    elif re.match(r'^(?:D|Red)istributable\b', section.licence):
        return DistState.non_free
    elif re.match(r'^GPL(?:v[23]|\+)?\b|^Dual GPL(?:v[23])?/', section.licence):
        return (DistState.free if is_source_available(section)
                else DistState.undistributable)
    else:
        # Unrecognised and probably undistributable
        return DistState.undistributable

def main(source_dir='.'):
    source_path = pathlib.Path(source_dir)
    added_path = pathlib.Path('debian/added-firmware')

    config = Config.read()
    with open("debian/copyright") as f:
        exclusions = deb822.Deb822(f).get("Files-Excluded", '').strip().split()
    link_exclusions = config.base.links_excluded
    unpackaged = config.base.files_unpackaged

    package_file_res = []
    for config_entry in config.package:
        package = config_entry.name
        package_file_res.append(
            ([pattern_to_re(pattern)
              for pattern in config_entry.files],
             [pattern_to_re(pattern)
              for pattern in config_entry.files_excluded])
        )

    for section in FirmwareWhence((source_path / 'WHENCE').open()):
        dist_state = check_section(section)

        for file_info in section.files.values():
            # will this file be included in the source package?
            if not any(fnmatch.fnmatch(file_info.binary, exclusion)
                       for exclusion in exclusions):
                if dist_state == DistState.non_free:
                    # Will it be included in any binary package?
                    if any(
                        (any(inc_re.fullmatch(file_info.binary)
                             for inc_re in inc_res)
                         and not any(exc_re.fullmatch(file_info.binary)
                                    for exc_re in exc_res))
                        for inc_res, exc_res in package_file_res
                    ):
                        update_file(source_path, added_path, file_info.binary)
                    elif any(fnmatch.fnmatch(file_info.binary, ignore_patt)
                             for ignore_patt in unpackaged):
                        pass
                    else:
                        print('I: %s is not included in any binary package' %
                              file_info.binary)
                elif dist_state == DistState.undistributable:
                    print('W: %s appears to be undistributable' %
                          file_info.binary)

        if dist_state == DistState.non_free:
            for link, target in section.links.items():
                link_path = pathlib.Path(link)
                target_path = ((link_path.parent / target)
                               .resolve()
                               .relative_to(pathlib.Path.cwd()))
                # - Is the target a file?
                # - Have either the target or the link itself been
                #   explicitly excluded?
                # - Will the link not be included in any binary package?
                if str(target_path) in section.files \
                   and not any(fnmatch.fnmatch(str(target_path), exclusion)
                               for exclusion in exclusions) \
                   and not any(fnmatch.fnmatch(link, exclusion)
                               for exclusion in link_exclusions) \
                   and not any(
                       (any(inc_re.fullmatch(link)
                            for inc_re in inc_res)
                        and not any(exc_re.fullmatch(link)
                                    for exc_re in exc_res))
                       for inc_res, exc_res in package_file_res
                   ):
                    print(f'I: {link} symlink is not included in any binary package')

def update_file(source_path, added_path, filename):
    source_file = source_path / filename
    over_file = added_path / filename
    if over_file.is_file():
        if not filecmp.cmp(source_file, over_file, True):
            print('I: %s: changed' % filename)

if __name__ == '__main__':
    main(*sys.argv[1:])
